"""
Tooling to help detect versioning issues with Reboot generated code.

NOTE: this module's exported interface MUST STAY BACKWARDS COMPATIBLE between
      releases; it is used by old generated Reboot code to detect versioning
      issues relative to newer Reboot libraries.
"""

from rebootdev.aio.exceptions import InputError
from rebootdev.version import REBOOT_VERSION


class IncompatibleVersionError(InputError):
    """
    Raised when generated code is incompatible with the current Reboot library
    version.
    """


def check_generated_code_compatible(
    generated_code_reboot_version: str
) -> None:
    """
    Checks if code generated by the given Reboot version is compatible

    If this library is too new to work with the given Reboot version's generated
    code, an error is raised.

    Args:
        generated_code_reboot_version: The Reboot version of the generated code.
    """
    if REBOOT_VERSION != generated_code_reboot_version:
        raise IncompatibleVersionError(
            f"This code was generated for Reboot version "
            f"'{generated_code_reboot_version}', but the installed Reboot "
            f"version is '{ REBOOT_VERSION }'. Please run `rbt dev run` or "
            "`rbt protoc` to regenerate this code."
        )
